{************************************************************************
  This Unit was modified to be interfaced in IBObjects
  by Lorenzo Mengoni - (Ancona, Italy)

  for suggestion and contributions email to lorenzo@mengoni.com

  This software is distributed on an "AS IS" basis,
  Under the GPL License from Free Software Foundation
  WITHOUT WARRANTY OF ANY KIND, either express or implied.
  If you live or are intended to use this software in countries
  where this is not applicable ("AS IS" basis), DO NOT USE IT.

                                                                         }
unit LMIBOQueryStatement;

interface

uses
  SysUtils, Classes, Db, IB_Components, IBODataset, IB_Header, Types, IB_Constants;

type
  TFieldRec = record
    Name:String;
    Select:Boolean;
    Computed:Boolean;
    Pk:Boolean;
    NotNull:Boolean;
    Length:Integer;
    FieldScale,
    FieldType : Integer;
    IntSubType:Integer;
    SubType:String;
    SegmentLength:Integer;
    CollationId:Integer;
    Collation:String;
    CharSetId:Integer;
    Column:String;
    CharacterSet:String;
    Domain:String;
    DefaultValue:String;
  end;

  TFieldArray = array of TFieldRec;

  TIndexDef = record
    Name        : String;
    Table       : String;
    IndexType   : integer;
    Unique      : integer;
    Inactive    : integer;
    Fields      : TStringDynArray;
    Description : String;
  end;

  TRefConstraintDef = record
     Name:String;                   
     Deferrable:String;             
     InitiallyDeferred:String;      
     Match_Option:String;           
     Update_Rule:String;            
     DeleteRule:String;
     IndexName:String;              
     RelationName:String;
     IndexDef:TIndexDef;
     rPK:TStringDynArray;
  end;

  TRefConstraintDefArray = array of TRefConstraintDef;

  TJoinField=record
     TableName:string;
     DisplayField:String;
     OnFKField:String;
     OnLocField:String;
     TableAlias:String;
     FieldAlias:String;
     LookUpSql:String;
  end;

  TJoinFieldArray = array of TJoinField;


  TSqlStatements = class(TObject)
  private
    FSelectSQL: TStringList;
    FDeleteSQL: TStringList;
    FUpdateSQL: TStringList;
    FInsertSQL: TStringList;
    fPk: TStringList;
    fFields: TFieldArray;
    FIB_Connection: TIB_Connection;
    FTableName: String;
    fView:boolean;
    fAssignedFields:boolean;
    FUse_OLD_In_Pk: boolean;
    FSelect_In_Pk_Order: boolean;
    fConstraints: TRefConstraintDefArray;
    FJoin_FK_Fields: boolean;
    fJoinFields:TJoinFieldArray;
    fSimulatePK:boolean;
    function GetDeleteSQL: TStrings;
    function GetInsertSQL: TStrings;
    function GetSelectSQL: TStrings;
    function GetUpdateSQL: TStrings;

    procedure SetDeleteSQL(const Value: TStrings);
    procedure SetInsertSQL(const Value: TStrings);
    procedure SetSelectSQL(const Value: TStrings);
    procedure SetUpdateSQL(const Value: TStrings);
    procedure SetIB_Connection(const Value: TIB_Connection);
    procedure SetTableName(const Value: String);

    procedure DoSelectSQL;
    procedure DoDeleteSQL;
    procedure DoUpdateSQL;
    procedure DoInsertSQL;

    procedure GetFieldLists;
    procedure SetView(const Value: boolean);
    function GetPK: TStrings;
    procedure SetUse_OLD_In_Pk(const Value: boolean);
    procedure SetSelect_In_Pk_Order(const Value: boolean);
    function PrintValidation(ToValidate: String; flag: Boolean): String;
    function GetCharacterSets(CharSetId, Collation: Short;
      CollateOnly: Boolean): String;
    function GetArrayField(FieldName: String): String;
    function GetCollation(CollationId: Short): String;
    procedure GetDomainData(const Domain: String; var ffield: TFieldRec);
    procedure GetReferentialConstList(
      var ObjectList: TRefConstraintDefArray; const TableName: string);
    procedure GetIndexData(var Data: TIndexDef; const IndexName: string);
    procedure GetPKList(const RelationName: String;
      var PK: TStringDynArray);
    procedure FreeConstraints;
    procedure SetJoin_FK_Fields(const Value: boolean);

  public
    constructor Create;
    destructor Destroy; override;
    procedure GenerateSQL;
    property IB_Connection:TIB_Connection read FIB_Connection write SetIB_Connection;
    property TableName:String read FTableName write SetTableName;
    property SelectSQL:TStrings read GetSelectSQL write SetSelectSQL;
    property DeleteSQL:TStrings read GetDeleteSQL write SetDeleteSQL;
    property UpdateSQL:TStrings read GetUpdateSQL write SetUpdateSQL;
    property InsertSQL:TStrings read GetInsertSQL write SetInsertSQL;
    procedure PrepareIBOQuery(q:TIBOQuery;const aTableName: String); overload;
    procedure PrepareIBOQuery(q:TIBOQuery;const aTableName: String; aFields:array of string); overload;
    procedure AssignPK(List:TStrings);
    procedure AssignFields(List:TStrings; const Computed:boolean=False);
    procedure AssignComputedFields(List:TStrings);
    procedure SelectField(aField:String; State:boolean);
    property View:boolean read FView write SetView;
    property Fields: TFieldArray read fFields;
    property PK: TStrings read GetPK;
    property Use_OLD_In_Pk:boolean read FUse_OLD_In_Pk write SetUse_OLD_In_Pk;
    property Select_In_Pk_Order:boolean read FSelect_In_Pk_Order write SetSelect_In_Pk_Order;
    function GetFieldData(const aName:String):TFieldRec;
    function GetFieldDataPK(const index: integer): TFieldRec;
    property Constraints: TRefConstraintDefArray read fConstraints;
    function GetDefaultValue(const index:integer):string;
    property Join_FK_Fields:boolean read FJoin_FK_Fields write SetJoin_FK_Fields;
    function QuoteId(const Value:String):String;
    function QuoteIdTF(const Table, Field: String): String;
    function HasJoin:boolean;
    property JoinFields:TJoinFieldArray read fJoinFields;
    function FieldHasJoin(const FieldName:string):boolean;

  end;

implementation

uses IBOAdminUtils, IB_Schema;

{ TSqlStatements }

procedure TSqlStatements.AssignComputedFields(List: TStrings);
var j:integer;
begin
  List.BeginUpdate;
  List.Clear;
  for j:=Low(fFields) to High(fFields) do
    if fFields[j].Computed then
       List.Add(fFields[j].Name);
  List.EndUpdate;
end;

procedure TSqlStatements.AssignFields(List: TStrings; const Computed:boolean=False);
var j:integer;
begin
  List.BeginUpdate;
  List.Clear;
  if Computed then
    for j:=Low(fFields) to High(fFields) do
      List.Add(fFields[j].Name)
  else
  for j:=Low(fFields) to High(fFields) do
    if not fFields[j].Computed then
       List.Add(fFields[j].Name);
  List.EndUpdate;
end;

procedure TSqlStatements.AssignPK(List: TStrings);
begin
  List.BeginUpdate;
  List.Clear;
  List.Assign(fPk);
  List.EndUpdate;
end;

constructor TSqlStatements.Create;
begin
  FSelectSQL:= TStringList.Create;
  FDeleteSQL:= TStringList.Create;
  FUpdateSQL:= TStringList.Create;
  FInsertSQL:= TStringList.Create;
  fPk:= TStringList.Create;
  FIB_Connection:=NIl;
  FTableName:='';
  FUse_OLD_In_Pk := True;
  FSelect_In_Pk_Order  := True;
  FJoin_FK_Fields := False;
  fSimulatePK := False;
end;

destructor TSqlStatements.Destroy;
begin
  FSelectSQL.Free;
  FDeleteSQL.Free;
  FUpdateSQL.Free;
  FInsertSQL.Free;
  fPk.Free;
  SetLength(fFields,0);
  SetLength(fJoinFields,0);
  FreeConstraints;
  inherited;
end;

procedure TSqlStatements.FreeConstraints;
var n:integer;
begin
  if Length(fConstraints) > 0 then
  try
    for n:= Low(fConstraints) to High(fConstraints) do
    try
      SetLength(fConstraints[n].rPK,0);
      SetLength(fConstraints[n].IndexDef.Fields,0);
    except
    end;
    SetLength(fConstraints,0);
  except
  end;
end;


procedure TSqlStatements.DoDeleteSQL;
var j:integer;
    s:String;
begin
   DeleteSQL.Clear;
   if not View then
   begin
     if fPk.Count>0 then
     begin
        DeleteSQL.Clear;
        DeleteSQL.Add('delete from '+QuoteId(TableName));
        DeleteSQL.Add('where');
        for j:=0 to Pred(fPk.Count) do
        begin
          s:=fPk[j];
          if FUse_OLD_In_Pk then
            s := '  '+QuoteId(s)+' = :'+QuoteId('OLD_'+s)
          else
            s := '  '+QuoteId(s)+' = :'+QuoteId(s);
          if j<Pred(fPk.Count) then s:=s+' and ';
          DeleteSQL.Add(s);
        end;
      end;
   end;
end;

procedure TSqlStatements.DoInsertSQL;
var j: integer;
    s: string;
    first:boolean;
begin
  InsertSql.Clear;
  if not View then
  begin
    InsertSql.Add('insert into '+QuoteId(TableName));
    InsertSql.Add('(');
    first:=true;
    for j:=Low(fFields) to High(fFields) do
      if fFields[j].select and not fFields[j].computed then
      begin
        if first then
        begin
          s:= '  '+QuoteId(fFields[j].Name);
          first:=False;
        end
        else
          s:= ', '+QuoteId(fFields[j].Name);
        InsertSql.Add(s);
      end;
    InsertSql.Add(')');
    InsertSql.Add('values');
    InsertSql.Add('(');
    first:=true;
    for j:=Low(fFields) to High(fFields) do
      if fFields[j].select and not fFields[j].computed then
      begin
        if first then
        begin
          s:= '  :'+QuoteId(fFields[j].Name);
          first:=False;
        end
        else
          s:= ', :'+QuoteId(fFields[j].Name);
        InsertSql.Add(s);
      end;
    InsertSql.Add(')');
  end;
end;

function FieldIsPresent(const FieldName:string; var JoinFields:TJoinFieldArray):boolean;
var
  j:integer;
begin
  Result:=False;
  for j:=Low(JoinFields) to High(JoinFields) do
    if JoinFields[j].DisplayField = FieldName then
    begin
      Result:=true;
      Break;
    end;
end;

procedure TSqlStatements.DoSelectSQL;
var j,x,y: integer;
    s: string;
    first:Boolean;
    addJoin:boolean;
    n:integer;
    fkf:TStringList;
begin
    addJoin := FJoin_FK_Fields and (Length(fConstraints)>0);
    SelectSql.Clear;
    first:=true;
    for j:=Low(fFields) to High(fFields) do
    begin
      if fFields[j].select then
      begin
      if first then
      begin
        first := False;
        SelectSql.Add('select '+QuoteIdTF(TableName,fFields[j].Name));
      end
      else
        SelectSql.Add('      ,'+QuoteIdTF(TableName,fFields[j].Name));
      end;
    end;
    if SelectSql.Text='' then
      SelectSql.Add('select *');
    if addJoin then
    begin
      fkf:=TStringList.Create;
      try
        for n:=Low(fConstraints) to High(fConstraints) do
        begin
          IB_schema.SchemaFieldNames(FIB_Connection,FIB_Connection.DefaultTransaction,fkf,fConstraints[n].RelationName);
          if fkf.Count>0 then
          for j:=Low(fConstraints[n].rPK) to High(fConstraints[n].rPK) do
          begin
            if fkf.IndexOf(fConstraints[n].rPK[j]) >=0 then
              fkf.Delete(fkf.IndexOf(fConstraints[n].rPK[j]));
          end;
          if fkf.Count>0 then
          begin
            x:=0;
            while (pos('ID_',fkf[x])=1) and (x<fkf.Count-1) do inc(x);
              SetLength(fJoinFields,Length(fJoinFields)+1);
              y:= High(fJoinFields);
              fJoinFields[y].TableAlias:= Char(High(fJoinFields)+65);
              fJoinFields[y].FieldAlias:= copy(StringReplace(fkf[x]+'_'+fConstraints[n].IndexDef.Fields[0]+'_J',' ','_',[rfReplaceAll]),1,31);
              fJoinFields[y].TableName := fConstraints[n].RelationName;
              fJoinFields[y].DisplayField := fkf[x];
              fJoinFields[y].OnFKField := fConstraints[n].rPK[0];
              fJoinFields[y].OnLocField := fConstraints[n].IndexDef.Fields[0];
              SelectSql.Add('  ,'+fJoinFields[y].TableAlias+'.'+
                                 QuoteId(fJoinFields[y].DisplayField)+' as '+
                                 fJoinFields[y].FieldAlias);
              fJoinFields[y].LookUpSql :=
              'select '+QuoteId(fJoinFields[y].OnFKField)+', '+QuoteId(fJoinFields[y].DisplayField)+
              ' from '+QuoteId(fJoinFields[y].TableName)+ ' order by '+QuoteId(fJoinFields[y].DisplayField);
          end;
        end;
      finally
        fkf.Free;
      end;
    end;
    SelectSql.Add('from '+QuoteId(TableName));
    if addJoin then
    begin
      for n:=Low(fJoinFields) to High(fJoinFields) do
      begin
        if fJoinFields[n].TableAlias = '' then
        begin
          SelectSql.Add('left outer join '+QuoteId(fJoinFields[n].TableName));
          SelectSql.Add('on '+QuoteIdTF(TableName,fJoinFields[n].OnLocField)+' = '+QuoteIdTF(fJoinFields[n].TableName,fJoinFields[n].OnFKField));
        end
        else
        begin
          SelectSql.Add('left outer join '+QuoteId(fJoinFields[n].TableName) +'  '+ fJoinFields[n].TableAlias);
          SelectSql.Add('on '+QuoteIdTF(TableName,fJoinFields[n].OnLocField)+' = '+fJoinFields[n].TableAlias+'.'+QuoteId(fJoinFields[n].OnFKField));
        end;
      end;
    end;
    if FSelect_In_Pk_Order and (fPk.Count>0) then
    begin
      SelectSql.Add('order by');
      for j:=0 to Pred(fPk.Count) do
      begin
        s:=QuoteIdTF(TableName,fPk[j]);
        if j<Pred(fPk.Count) then
          s:=s+', ';
        SelectSql.Add(s);
      end;
    end;
end;

procedure TSqlStatements.DoUpdateSQL;
var j: integer;
    s: string;
    first:boolean;
begin
  UpdateSql.Clear;
  if not view then
  begin
    if fPk.Count>0 then
    begin
      UpdateSql.Add('update '+QuoteId(TableName));
      UpdateSql.Add('set');
      first:=True;
      for j:=Low(fFields) to High(fFields) do
      if FUse_OLD_In_Pk then
      begin
        if fFields[j].select and not fFields[j].Computed then
          begin
            s:=fFields[j].Name;
            if first then
            begin
              s := ' '+QuoteId(s)+' = :'+QuoteId(s);
              first:=False;
            end
            else
              s := ', '+QuoteId(s)+' = :'+QuoteId(s);
            UpdateSql.Add(s);
          end;
      end
      else
      begin
        if fFields[j].select and not fFields[j].Computed and not fFields[j].pk then
          begin
            s:=fFields[j].Name;
            if first then
            begin
              s := ' '+QuoteId(s)+' = :'+QuoteId(s);
              first:=False;
            end
            else
              s := ', '+QuoteId(s)+' = :'+QuoteId(s);
            UpdateSql.Add(s);
          end;
      end;
      UpdateSql.Add('where');
      for j:=0 to Pred(fPk.Count) do
      begin
        s:=fPk[j];
        if FUse_OLD_In_Pk then
          s := '  '+QuoteId(s)+' = :'+QuoteId('OLD_'+s)
        else
          s := '  '+QuoteId(s)+' = :'+QuoteId(s);
        if j<Pred(fPk.Count) then
          s:=s+' and ';
        UpdateSql.Add(s);
      end;
    end;
  end;
end;

procedure TSqlStatements.GenerateSQL;
begin
  DoSelectSQL;
  DoDeleteSQL;
  DoUpdateSQL;
  DoInsertSQL;
end;

function TSqlStatements.PrintValidation(ToValidate: String;
  flag: Boolean): String;
var
  IsSQL : Boolean;
begin
  IsSql := false;

  Result := '';
  ToValidate := Trim(ToValidate);

  if flag then
  begin
    if ToValidate[1] = '(' then
      IsSQL := true;
  end
  else
    if (Pos(ToValidate, 'check') = 1) or (Pos(ToValidate, 'CHECK') = 1) then
      IsSQL := TRUE;

	if not IsSQL then
  begin
    if Flag then
      Result := Result + '/* ' + ToValidate + ' */'
    else
      Result := Result + '(' + ToValidate + ')';
  end
  else
    Result := ToValidate;
end;

const
  CollationSQL =
    'SELECT CST.RDB$CHARACTER_SET_NAME '+
    '     , COL.RDB$COLLATION_NAME '+
    '     , CST.RDB$DEFAULT_COLLATE_NAME ' +
    'FROM RDB$COLLATIONS COL '+
    'JOIN RDB$CHARACTER_SETS CST ON ' +
    '  COL.RDB$CHARACTER_SET_ID = CST.RDB$CHARACTER_SET_ID ' +
    'WHERE ' +
    '  COL.RDB$COLLATION_ID = :COLLATION_ID AND ' +
    '  CST.RDB$CHARACTER_SET_ID = :CHAR_SET_ID ' +
    'ORDER BY COL.RDB$COLLATION_NAME, CST.RDB$CHARACTER_SET_NAME';

  NonCollationSQL =
    'SELECT CST.RDB$CHARACTER_SET_NAME ' +
    'FROM RDB$CHARACTER_SETS CST ' +
    'WHERE CST.RDB$CHARACTER_SET_ID = :CHARSETID ' +
    'ORDER BY CST.RDB$CHARACTER_SET_NAME';

  ArraySQL =
    'SELECT * FROM RDB$FIELD_DIMENSIONS FDIM ' +
    'WHERE ' +
    '  FDIM.RDB$FIELD_NAME = :FIELDNAME ' +
    'ORDER BY FDIM.RDB$DIMENSION';


  CollationIdSQL =
    'SELECT '+
    '      COL.RDB$COLLATION_NAME '+
    '      ' +
    'FROM RDB$COLLATIONS COL '+
    'WHERE ' +
    '  COL.RDB$COLLATION_ID = :COLLATION_ID  ';



function TSqlStatements.GetCharacterSets(CharSetId, Collation: Short;
  CollateOnly: Boolean): String;
var
  CharSetSQL : TIB_Cursor;
begin
  CharSetSQL := TIB_Cursor.Create(FIB_Connection);
  CharSetSQL.IB_Connection := FIB_Connection;
  CharSetSQL.IB_Transaction := FIB_Connection.DefaultTransaction;

  try
    if Collation <> 0 then
    begin
      CharSetSQL.SQL.Text := CollationSQL;
      CharSetSQL.Prepare;
      CharSetSQL.Params.ByName('CHAR_SET_ID').AsInteger := CharSetId;
      CharSetSQL.Params.ByName('COLLATION_ID').AsInteger := Collation;
      CharSetSQL.Open;

      { Is specified collation the default collation for character set? }
      if (Trim(CharSetSQL.FieldByName('RDB$DEFAULT_COLLATE_NAME').AsString) =
         Trim(CharSetSQL.FieldByName('RDB$COLLATION_NAME').AsString)) then
      begin
        if not CollateOnly then
          Result := ' CHARACTER SET ' + Trim(CharSetSQL.FieldByName('RDB$CHARACTER_SET_NAME').AsString);
      end
      else
        if CollateOnly then
          Result := ' COLLATE ' + Trim(CharSetSQL.FieldByName('RDB$COLLATION_NAME').AsString)
        else
          Result := ' CHARACTER SET ' +
            Trim(CharSetSQL.FieldByName('RDB$CHARACTER_SET_NAME').AsString) +
            ' COLLATE ' +
            Trim(CharSetSQL.FieldByName('RDB$COLLATION_NAME').AsString);
    end
    else
      if CharSetId <> 0 then
      begin
        CharSetSQL.SQL.Text := NonCollationSQL;
        CharSetSQL.Prepare;
        CharSetSQL.Params.ByName('CHARSETID').AsSmallInt := CharSetId;
        CharSetSQL.Open;
        Result := ' CHARACTER SET ' + Trim(CharSetSQL.FieldByName('RDB$CHARACTER_SET_NAME').AsString);
      end;
  finally
    CharSetSQL.Free;
  end;
end;

function TSqlStatements.GetCollation(CollationId: Short): String;
var
  CharSetSQL : TIB_Cursor;
begin
  result := '';
  CharSetSQL := TIB_Cursor.Create(FIB_Connection);
  CharSetSQL.IB_Connection := FIB_Connection;
  CharSetSQL.IB_Transaction := FIB_Connection.DefaultTransaction;

  try
    if CollationId <> 0 then
    begin
      CharSetSQL.SQL.Text := CollationIdSQL;
      CharSetSQL.Prepare;
      CharSetSQL.Params.ByName('COLLATION_ID').AsInteger := CollationId ;
      CharSetSQL.Open;
      Result := Trim(CharSetSQL.FieldByName('RDB$COLLATION_NAME').AsString)
    end
  finally
    CharSetSQL.Free;
  end;
end;


function TSqlStatements.GetArrayField(FieldName: String): String;
var
  qryArray : TIB_Cursor;
begin
  qryArray := TIB_Cursor.Create(FIB_Connection);
  qryArray.IB_Connection := FIB_Connection;
  qryArray.IB_Transaction := FIB_Connection.DefaultTransaction;
  Result := '[';
  qryArray.SQL.Text := ArraySQL;
  qryArray.Prepare;
  qryArray.Params.ByName('FIELDNAME').AsString := FieldName;
  qryArray.Open;

    {  Format is [lower:upper, lower:upper,..]  }

  while not qryArray.Eof do
  begin
    if (qryArray.FieldByName('RDB$DIMENSION').AsInteger > 0) then
      Result := Result + ', ';
    Result := Result + qryArray.FieldByName('RDB$LOWER_BOUND').AsString + ':' +
           qryArray.FieldByName('RDB$UPPER_BOUND').AsString;
    qryArray.Next;
  end;

  Result := Result + '] ';
  qryArray.Free;
  
end;



function TSqlStatements.GetDeleteSQL: TStrings;
begin
  Result := fDeleteSQL;
end;

const
  TableListSQL1 =
    'SELECT '+ //'* '+
    ' RDB$RELATION_FIELDS.RDB$FIELD_NAME '+
    ', RDB$RELATION_FIELDS.RDB$FIELD_POSITION '+
    ', RDB$RELATIONS.RDB$OWNER_NAME '+
    ', RDB$RELATIONS.RDB$EXTERNAL_FILE '+
    ', RDB$FIELDS.RDB$COMPUTED_BLR '+
    ', RDB$FIELDS.RDB$COMPUTED_SOURCE '+
    ', RDB$FIELDS.RDB$FIELD_TYPE '+
    ', RDB$FIELDS.RDB$FIELD_SCALE '+
    ', RDB$FIELDS.RDB$FIELD_NAME '+
    ', RDB$FIELDS.RDB$SYSTEM_FLAG '+
    ', RDB$FIELDS.RDB$DEFAULT_SOURCE '+
    ', RDB$RELATION_FIELDS.RDB$NULL_FLAG '+
    ', RDB$RELATIONS.RDB$SYSTEM_FLAG '+
    ', RDB$FIELDS.RDB$COLLATION_ID '+
    ', RDB$FIELDS.RDB$CHARACTER_SET_ID '+
    ', RDB$RELATION_FIELDS.RDB$RELATION_NAME '+
    ', RDB$FIELDS.RDB$CHARACTER_LENGTH '+
    ', RDB$FIELDS.RDB$DIMENSIONS '+
    ', RDB$FIELDS.RDB$FIELD_SUB_TYPE '+
    ', RDB$FIELDS.RDB$FIELD_SCALE '+
    ', RDB$FIELDS.RDB$FIELD_LENGTH '+
    ', RDB$FIELDS.RDB$SEGMENT_LENGTH '+
    ', RDB$RELATION_FIELDS.RDB$COLLATION_ID '+
    ' FROM RDB$RELATIONS '+
    ' JOIN RDB$RELATION_FIELDS ON RDB$RELATION_FIELDS.RDB$RELATION_NAME = RDB$RELATIONS.RDB$RELATION_NAME '+
    ' JOIN RDB$FIELDS ON RDB$RELATION_FIELDS.RDB$FIELD_SOURCE = RDB$FIELDS.RDB$FIELD_NAME ' +
    ' WHERE RDB$RELATION_FIELDS.RDB$RELATION_NAME = :TableName ' +
    ' ORDER BY RDB$RELATION_FIELDS.RDB$FIELD_POSITION, RDB$RELATION_FIELDS.RDB$FIELD_NAME';


  TableListSQL3 =
    'SELECT '+ //'* '+
    ' RDB$RELATION_FIELDS.RDB$FIELD_NAME '+
    ', RDB$RELATION_FIELDS.RDB$FIELD_POSITION '+
    ', RDB$RELATIONS.RDB$OWNER_NAME '+
    ', RDB$RELATIONS.RDB$EXTERNAL_FILE '+
    ', RDB$FIELDS.RDB$COMPUTED_BLR '+
    ', RDB$FIELDS.RDB$COMPUTED_SOURCE '+
    ', RDB$FIELDS.RDB$FIELD_TYPE '+
    ', RDB$FIELDS.RDB$FIELD_SCALE '+
    ', RDB$FIELDS.RDB$FIELD_NAME '+
    ', RDB$FIELDS.RDB$SYSTEM_FLAG '+
    ', RDB$FIELDS.RDB$DEFAULT_SOURCE '+
    ', RDB$RELATION_FIELDS.RDB$NULL_FLAG '+
    ', RDB$RELATIONS.RDB$SYSTEM_FLAG '+
    ', RDB$FIELDS.RDB$COLLATION_ID '+
    ', RDB$FIELDS.RDB$CHARACTER_SET_ID '+
    ', RDB$RELATION_FIELDS.RDB$RELATION_NAME '+
    ', RDB$FIELDS.RDB$CHARACTER_LENGTH '+
    ', RDB$FIELDS.RDB$DIMENSIONS '+
    ', RDB$FIELDS.RDB$FIELD_SUB_TYPE '+
    ', RDB$FIELDS.RDB$FIELD_PRECISION '+
    ', RDB$FIELDS.RDB$FIELD_SCALE '+
    ', RDB$FIELDS.RDB$FIELD_LENGTH '+
    ', RDB$FIELDS.RDB$SEGMENT_LENGTH '+
    ', RDB$RELATION_FIELDS.RDB$COLLATION_ID '+
    ' FROM RDB$RELATIONS '+
    ' JOIN RDB$RELATION_FIELDS ON RDB$RELATION_FIELDS.RDB$RELATION_NAME = RDB$RELATIONS.RDB$RELATION_NAME '+
    ' JOIN RDB$FIELDS ON RDB$RELATION_FIELDS.RDB$FIELD_SOURCE = RDB$FIELDS.RDB$FIELD_NAME ' +
    ' WHERE RDB$RELATION_FIELDS.RDB$RELATION_NAME = :TableName ' +
    ' ORDER BY RDB$RELATION_FIELDS.RDB$FIELD_POSITION, RDB$RELATION_FIELDS.RDB$FIELD_NAME';


const SqlFields =
      'select  rf.rdb$field_name, '+
      '        /*rf.rdb$relation_name,*/ '+
      '        /*rf.rdb$field_source,*/ '+
      '        f.rdb$computed_source '+
      'from    rdb$relation_fields rf '+
      'left outer join rdb$fields f on f.rdb$field_name = rf.rdb$field_source '+
      'where   rf.rdb$relation_name=?TableName '+
      'order by rf.rdb$field_position;';

      SqlPK =
      'select i.rdb$field_name                                  '+
      'from    rdb$relation_constraints r, rdb$index_segments i '+
      'where   r.rdb$relation_name=?TableName and               '+
      '        r.rdb$constraint_type=''PRIMARY KEY'' and        '+
      '        r.rdb$index_name=i.rdb$index_name                '+
      'order by i.rdb$field_position;                           ';

      SqlUIDX =
      ' select i.rdb$index_name                          '+
      ' from rdb$indices i                               '+
      ' where i.rdb$relation_name=?TableName             '+
      ' and i.rdb$unique_flag=1                          '+
      ' and i.rdb$index_name not in (                    '+
      '   select distinct r.rdb$index_name               '+
      '   from rdb$relation_constraints r                '+
      '   where r.rdb$relation_name=i.rdb$relation_name  '+
      '   and r.rdb$index_name is not null               '+
      '  )                                               ';

      SqlFldIDX =
      ' select  i.rdb$field_name                '+
      ' from    rdb$index_segments i            '+
      ' where   i.rdb$index_name = :index_name  '+
      ' order by i.rdb$field_position;          ';

  PrecisionSQL =
    'SELECT * FROM RDB$FIELDS ' +
    'WHERE RDB$FIELD_NAME = :FIELDNAME';

  ConstraintSQL =
    'SELECT RCO.RDB$CONSTRAINT_NAME, RDB$CONSTRAINT_TYPE, RDB$RELATION_NAME, ' +
    'RDB$DEFERRABLE, RDB$INITIALLY_DEFERRED, RDB$INDEX_NAME, RDB$TRIGGER_NAME ' +
    'FROM RDB$RELATION_CONSTRAINTS RCO, RDB$CHECK_CONSTRAINTS CON ' +
    'WHERE ' +
    '  CON.RDB$TRIGGER_NAME = :FIELDNAME AND ' +
    '  CON.RDB$CONSTRAINT_NAME = RCO.RDB$CONSTRAINT_NAME AND ' +
    '  RCO.RDB$CONSTRAINT_TYPE = ''NOT NULL'' AND ' +
    '  RCO.RDB$RELATION_NAME = :RELATIONNAME';


type
  TSQLType = record
    SqlType : Integer;
    TypeName : String;
  end;

  TSQLTypes = Array[0..13] of TSQLType;

const
  MAX_INTSUBTYPES = 2;
  MAXSUBTYPES = 8;     { Top of subtypes array }

  ColumnTypes : TSQLTypes = (
    (SqlType : blr_short; TypeName :	'SMALLINT'),		{ NTX: keyword }
    (SqlType : blr_long; TypeName : 'INTEGER'),		{ NTX: keyword }
    (SqlType : blr_quad; TypeName : 'QUAD'),		{ NTX: keyword }
    (SqlType : blr_float; TypeName : 'FLOAT'),		{ NTX: keyword }
    (SqlType : blr_text; TypeName : 'CHAR'),		{ NTX: keyword }
    (SqlType : blr_double; TypeName : 'DOUBLE PRECISION'),	{ NTX: keyword }
    (SqlType : blr_varying; TypeName : 'VARCHAR'),		{ NTX: keyword }
    (SqlType : blr_cstring; TypeName : 'CSTRING'),		{ NTX: keyword }
    (SqlType : blr_blob_id; TypeName : 'BLOB_ID'),		{ NTX: keyword }
    (SqlType : blr_blob; TypeName : 'BLOB'),		{ NTX: keyword }
    (SqlType : blr_sql_time; TypeName : 'TIME'),		{ NTX: keyword }
    (SqlType : blr_sql_date; TypeName : 'DATE'),		{ NTX: keyword }
    (SqlType : blr_timestamp; TypeName : 'TIMESTAMP'),		{ NTX: keyword }
    (SqlType : blr_int64; TypeName : 'INT64'));

  IntegralSubtypes : Array[0..2] of String = (
    'UNKNOWN',			{ Defined type, NTX: keyword }
    'NUMERIC',			{ NUMERIC, NTX: keyword }
    'DECIMAL');   		{ DECIMAL, NTX: keyword }

  BlobSubTypes : Array[0..8] of String = (
    'UNKNOWN',			 { NTX: keyword }
    'TEXT',			 { NTX: keyword }
    'BLR',			 { NTX: keyword }
    'ACL',			 { NTX: keyword }
    'RANGES',			 { NTX: keyword }
    'SUMMARY',			 { NTX: keyword }
    'FORMAT',			 { NTX: keyword }
    'TRANSACTION_DESCRIPTION',	 { NTX: keyword }
    'EXTERNAL_FILE_DESCRIPTION');{ NTX: keyword }


const
  DomainByNameSQL =
    'SELECT * FROM RDB$FIELDS FLD ' +
    'WHERE FLD.RDB$FIELD_NAME = :DomainName ' +
    'ORDER BY FLD.RDB$FIELD_NAME';


procedure TSqlStatements.GetDomainData(const Domain: String; var ffield:TFieldRec);
var
  i, SubType : Integer;
  PrecisionKnown : Boolean;
  qryDomains : TIB_Cursor;

begin
  qryDomains := TIB_Cursor.Create(nil);
  try
    qryDomains.IB_Connection := FIB_Connection;
    qryDomains.IB_Transaction := FIB_Connection.DefaultTransaction;
    qryDomains.SQL.Text := DomainByNameSQL;
    qryDomains.Prepare;
    qryDomains.Params[0].AsString := Domain;
    qryDomains.Open;
    ffield.Column := ffield.Column + ' /* ';
    for i := Low(ColumnTypes) to High(ColumnTypes) do
    begin
      if qryDomains.FieldByName('RDB$FIELD_TYPE').AsInteger = ColumnTypes[i].SQLType then
      begin
        ffield.FieldType := qryDomains.FieldByName('RDB$FIELD_TYPE').AsInteger;
        PrecisionKnown := FALSE;
        if FIB_Connection.Characteristics.dbODS_Version >= 10 then
        begin
          if qryDomains.FieldByName('RDB$FIELD_TYPE').AsInteger in [blr_short, blr_long, blr_int64] then
          begin
            { We are ODS >= 10 and could be any Dialect }
            if (FIB_Connection.SQLDialect >= 3) and
               (not qryDomains.FieldByName('RDB$FIELD_PRECISION').IsNull) and
               (qryDomains.FieldByName('RDB$FIELD_SUB_TYPE').AsInteger > 0) and
               (qryDomains.FieldByName('RDB$FIELD_SUB_TYPE').AsInteger <= MAX_INTSUBTYPES) then
            begin
              ffield.FieldScale := qryDomains.FieldByName('RDB$FIELD_SCALE').AsInteger;
              ffield.Column := ffield.Column + Format(' %s(%d, %d) ', [
                IntegralSubtypes [qryDomains.FieldByName('RDB$FIELD_SUB_TYPE').AsInteger],
                qryDomains.FieldByName('RDB$FIELD_PRECISION').AsInteger,
                -1 * qryDomains.FieldByName('RDB$FIELD_SCALE').AsInteger]);
              PrecisionKnown := true;
            end;
          end;
        end;
        if PrecisionKnown = false then
        begin
          { Take a stab at numerics and decimals }
          if (qryDomains.FieldByName('RDB$FIELD_TYPE').AsInteger = blr_short) and
              (qryDomains.FieldByName('RDB$FIELD_SCALE').AsInteger < 0) then
            ffield.Column := ffield.Column + Format(' NUMERIC(4, %d)',
              [-qryDomains.FieldByName('RDB$FIELD_SCALE').AsInteger] )
          else
            if (qryDomains.FieldByName('RDB$FIELD_TYPE').AsInteger = blr_long) and
                (qryDomains.FieldByName('RDB$FIELD_SCALE').AsInteger < 0) then
              ffield.Column  := ffield.Column  + Format(' NUMERIC(9, %d)',
                [-qryDomains.FieldByName('RDB$FIELD_SCALE').AsInteger] )
            else
              if (qryDomains.FieldByName('RDB$FIELD_TYPE').AsInteger = blr_double) and
                  (qryDomains.FieldByName('RDB$FIELD_SCALE').AsInteger  < 0) then
                ffield.Column := ffield.Column + Format(' NUMERIC(15, %d)',
                  [-qryDomains.FieldByName('RDB$FIELD_SCALE').AsInteger] )
              else
                ffield.Column := ffield.Column + ColumnTypes[i].TypeName;
        end;
        break;
      end;
    end;

    if qryDomains.FieldByName('RDB$FIELD_TYPE').AsInteger = blr_blob then
    begin
      subtype := qryDomains.FieldByName('RDB$FIELD_SUB_TYPE').AsInteger;
      ffield.Column := ffield.Column + ' SUB_TYPE ';
      ffield.SegmentLength := qryDomains.FieldByName('RDB$SEGMENT_LENGTH').AsInteger;
      if (subtype > 0) and (subtype <= MAXSUBTYPES) then
        ffield.SubType := blobSubTypes[subtype]
      else
        ffield.SubType := Format('%d', [subtype]);
      ffield.Column := ffield.Column + ffield.SubType;
      ffield.Column := ffield.Column + Format(' SEGMENT SIZE %d ', [ffield.SegmentLength]);
    end //end_if
    else
    if (qryDomains.FieldByName('RDB$FIELD_TYPE').AsInteger in [blr_text, blr_varying]) and
       (not qryDomains.FieldByName('RDB$CHARACTER_LENGTH').IsNull) then
    begin
      ffield.Length := qryDomains.FieldByName('RDB$FIELD_LENGTH').AsInteger;
      ffield.Column := ffield.Column + Format('(%d)', [qryDomains.FieldByName('RDB$FIELD_LENGTH').AsInteger]);
    end;

    { since the character set is part of the field type, display that
     information now. }
    if not qryDomains.FieldByName('RDB$CHARACTER_SET_ID').IsNull then
      ffield.Column := ffield.Column + GetCharacterSets(qryDomains.FieldByName('RDB$CHARACTER_SET_ID').AsInteger,
         0, FALSE);
    if not qryDomains.FieldByName('RDB$DIMENSIONS').IsNull then
      ffield.Column := ffield.Column + GetArrayField(ffield.Name);

    if not qryDomains.FieldByName('RDB$DEFAULT_SOURCE').IsNull then
    begin
      ffield.DefaultValue := Trim(qryDomains.FieldByName('RDB$DEFAULT_SOURCE').AsString);
      ffield.Column := ffield.Column + Format('%s%s %s', ['', '', ffield.DefaultValue]);
      if pos('default',ffield.DefaultValue)=1 then
        delete(ffield.DefaultValue,1,Length('default'));
      ffield.DefaultValue := Trim(ffield.DefaultValue);
      if ffield.DefaultValue<>'' then
      if (ffield.DefaultValue[1]= '''') and  (ffield.DefaultValue[Length(ffield.DefaultValue)] =  '''') then
      begin
        delete(ffield.DefaultValue,1,1);
        delete(ffield.DefaultValue,Length(ffield.DefaultValue),1);
      end;
    end;

    if not qryDomains.FieldByName('RDB$VALIDATION_SOURCE').IsNull then
      if Pos('CHECK', AnsiUpperCase(qryDomains.FieldByName('RDB$VALIDATION_SOURCE').AsString)) = 1 then
        ffield.Column := ffield.Column + Format('%s%s %s', ['', '',
           qryDomains.FieldByName('RDB$VALIDATION_SOURCE').AsString])
      else
        ffield.Column := ffield.Column + Format('%s%s /* %s */', ['', '',
           qryDomains.FieldByName('RDB$VALIDATION_SOURCE').AsString]);

    if qryDomains.FieldByName('RDB$NULL_FLAG').AsInteger = 1 then
      ffield.Column := ffield.Column + ' NOT NULL';

    { Show the collation order if one has been specified.  If the collation
       order is the default for the character set being used, then no collation
       order will be shown ( because it isn't needed ).

       If the collation id is 0, then the default for the character set is
       being used so there is no need to retrieve the collation information.}

    if (not qryDomains.FieldByName('RDB$COLLATION_ID').IsNull) and
       (qryDomains.FieldByName('RDB$COLLATION_ID').AsInteger <> 0) then
      ffield.Column := ffield.Column + GetCharacterSets(qryDomains.FieldByName('RDB$CHARACTER_SET_ID').AsInteger,
        qryDomains.FieldByName('RDB$COLLATION_ID').AsInteger, true);

    ffield.Column := ffield.Column + ' */';
  finally
    qryDomains.Free;
  end;
end;


procedure TSqlStatements.GetFieldLists;
var c:TIB_Cursor;
    t:TIB_Transaction;
    qryPrecision:TIB_Cursor;
    qryConstraints:TIB_Cursor;

    procedure AssignField(n:integer);
    var i:integer;
        PrecisionKnown:boolean;
        intchar:Integer;
    begin
      ffields[n].Name := c.FieldByName('RDB$RELATION_FIELDS.RDB$FIELD_NAME').AsString;
      ffields[n].Computed := c.FieldByName('RDB$FIELDS.RDB$COMPUTED_SOURCE').AsString<>'';
      ffields[n].Select := true;
      ffields[n].Pk:=False;
      ffields[n].NotNull:=False;
      ffields[n].Length:=0;
      ffields[n].FieldScale:=0;
      ffields[n].FieldType:=0;
      ffields[n].IntSubType:=0;
      ffields[n].SubType:= '';
      ffields[n].SegmentLength:=0;
      ffields[n].CollationId:=0;
      ffields[n].Collation:= '';
      ffields[n].CharSetId:=0;
      ffields[n].Column:= '';
      ffields[n].CharacterSet:= '';
      ffields[n].Domain:= '';
      ffields[n].DefaultValue:= '';

      intchar := 0;
      ffields[n].Column := QuoteIdentifier(FIB_Connection.SQLDialect, ffields[n].Name)+' ';

    {  Check first for computed fields, then domains.
       If this is a known domain, then just print the domain rather than type
       Domains won't have length, array, or blob definitions, but they
       may have not null, default and check overriding their definitions }

      if not c.FieldByName('RDB$COMPUTED_BLR').IsNull then
      begin
        ffields[n].Column := ffields[n].Column + ' COMPUTED BY ';
       if not c.FieldByName('RDB$COMPUTED_SOURCE').IsNull then
         ffields[n].Column := ffields[n].Column + PrintValidation(Trim(c.FieldByName('RDB$COMPUTED_SOURCE').AsString), true);
      end
      else
      begin
        ffields[n].FieldType := c.FieldByName('RDB$FIELD_TYPE').AsInteger;
        ffields[n].FieldScale := c.FieldByName('RDB$FIELD_SCALE').AsInteger;
        if not ((Copy(c.FieldByName('RDB$FIELDS.RDB$FIELD_NAME').AsString, 1, 4) = 'RDB$') and
          (c.FieldByName('RDB$FIELDS.RDB$FIELD_NAME').AsString[5] in ['0'..'9'])) and
          (c.FieldByName('RDB$SYSTEM_FLAG').AsInteger <> 1) then
        begin  //domain !!
          ffields[n].Domain := Trim(c.FieldByName('RDB$FIELDS.RDB$FIELD_NAME').AsString);
          ffields[n].Column := ffields[n].Column + QuoteIdentifier(FIB_Connection.SQLDialect, ffields[n].Domain);
          { International character sets }
          if (c.FieldByName('RDB$FIELD_TYPE').AsInteger in [blr_text, blr_varying])
              and (not c.FieldByName('RDB$COLLATION_ID').IsNull)
              and (c.FieldByName('RDB$COLLATION_ID').AsSmallint <> 0) then
          begin
            ffields[n].CollationId := c.FieldByName('RDB$COLLATION_ID').AsSmallInt;
            ffields[n].Collation := GetCollation(ffields[n].CollationId);
            ffields[n].CharSetId := c.FieldByName('RDB$CHARACTER_SET_ID').AsSmallInt;
            ffields[n].CharacterSet := GetCharacterSets(ffields[n].CharSetId,ffields[n].CollationId, true);
            ffields[n].Column := ffields[n].Column + ffields[n].CharacterSet;
          end;
          GetDomainData(ffields[n].Domain ,ffields[n]);
        end
        else
        begin
  	      { Look through types array }
          for i := Low(Columntypes) to High(ColumnTypes) do
          begin
            PrecisionKnown := false;
            if ffields[n].FieldType = ColumnTypes[i].SQLType then
            begin
              if FIB_Connection.Characteristics.dbODS_Version >= 10 then
              begin
                { Handle Integral subtypes NUMERIC and DECIMAL }
                if ffields[n].FieldType in
                        [blr_short, blr_long, blr_int64] then
                begin
                  { We are ODS >= 10 and could be any Dialect }
                  if FIB_Connection.SQLDialect = 1 then
                  begin
                    qryPrecision.Params.ByName('FIELDNAME').AsString := ffields[n].Name;                    qryPrecision.Open;
                    if not qryPrecision.FieldByName('RDB$FIELD_PRECISION').IsNull then
                    begin
                    { We are Dialect >=3 since FIELD_PRECISION is non-NULL }
                      if (qryPrecision.FieldByName('RDB$FIELD_SUB_TYPE').AsInteger > 0) and
                         (qryPrecision.FieldByName('RDB$FIELD_SUB_TYPE').AsInteger <= MAX_INTSUBTYPES) then
                      begin
                        ffields[n].Column := ffields[n].Column + Format('%s(%d, %d)',
                           [IntegralSubtypes[qryPrecision.FieldByName('RDB$FIELD_SUB_TYPE').AsInteger],
                           qryPrecision.FieldByName('RDB$FIELD_PRECISION').AsInteger,
                          -qryPrecision.FieldByName('RDB$FIELD_SCALE').AsInteger]);
                        PrecisionKnown := TRUE;
                      end;
                    end;
                    qryPrecision.Close;
                  end;
                end;
              end;

              if PrecisionKnown = FALSE then
              begin
                { Take a stab at numerics and decimals }
                if (ffields[n].FieldType = blr_short) and (ffields[n].FieldScale < 0) then
                  ffields[n].Column := ffields[n].Column + Format('NUMERIC(4, %d)', [-ffields[n].FieldScale])
                else
                  if (ffields[n].FieldType = blr_long) and (ffields[n].FieldScale < 0) then
                    ffields[n].Column := ffields[n].Column + Format('NUMERIC(9, %d)', [-ffields[n].FieldScale])
                  else
                    if (ffields[n].FieldType = blr_double) and (ffields[n].FieldScale < 0) then
                      ffields[n].Column := ffields[n].Column + Format('NUMERIC(15, %d)', [-ffields[n].FieldScale])
                    else
                      ffields[n].Column := ffields[n].Column + ColumnTypes[i].TypeName;
              end;
            end;
          end;
          if ffields[n].FieldType in [blr_text, blr_varying] then
          begin
            if c.FieldByName('RDB$FIELDS.RDB$CHARACTER_LENGTH').IsNull then
            begin
              ffields[n].Length := c.FieldByName('RDB$FIELDS.RDB$FIELD_LENGTH').AsInteger;
            end
            else
            begin
              ffields[n].Length := c.FieldByName('RDB$FIELDS.RDB$CHARACTER_LENGTH').AsInteger;
            end;
            ffields[n].Column := ffields[n].Column + Format('(%d)', [ffields[n].Length]);
          end;
            // ffields[n].Column := ffields[n].Column + Format('(%d)', [c.FieldByName('RDB$FIELDS.RDB$SEGMENT_LENGTH').AsInteger]);

          { Catch arrays after printing the type  }

          if not c.FieldByName('RDB$DIMENSIONS').IsNull then
            ffields[n].Column := ffields[n].Column + GetArrayField(c.FieldByName('RDB$RELATION_FIELDS.RDB$FIELD_NAME').AsString);

          if ffields[n].FieldType = blr_blob then
          begin
            ffields[n].IntSubType := c.FieldByName('RDB$FIELD_SUB_TYPE').AsSmallInt;
            ffields[n].SegmentLength := c.FieldByName('RDB$SEGMENT_LENGTH').AsInteger;

            if ffields[n].IntSubType in [0..MAXSUBTYPES] then
              ffields[n].SubType := BlobSubTypes[ffields[n].IntSubType]
            else
              ffields[n].SubType := IntToStr(ffields[n].IntSubType);
            ffields[n].Column := ffields[n].Column + ' SUB_TYPE ' + ffields[n].SubType +
                                 Format(' SEGMENT SIZE %d',[ffields[n].SegmentLength]);
          end;

          { International character sets }
          if ((ffields[n].FieldType in [blr_text, blr_varying]) or
              (ffields[n].FieldType = blr_blob)) and
             (not c.FieldByName('RDB$CHARACTER_SET_ID').IsNull) and
             (c.FieldByName('RDB$CHARACTER_SET_ID').AsInteger <> 0) then
          begin
            { Override rdb$fields id with relation_fields if present }
            ffields[n].CharSetId := 0;
            if not c.FieldByName('RDB$CHARACTER_SET_ID').IsNull then
              ffields[n].CharSetId := c.FieldByName('RDB$CHARACTER_SET_ID').AsInteger;
            ffields[n].CharacterSet := GetCharacterSets(ffields[n].CharSetId, 0, false);
            ffields[n].Column := ffields[n].Column + ffields[n].CharacterSet;
            intchar := 1;
          end;
        end;

        { Handle defaults for columns }
        { Originally This called PrintMetadataTextBlob,
            should no longer need }
        if not c.FieldByName('RDB$DEFAULT_SOURCE').IsNull then
        begin
          ffields[n].DefaultValue := Trim(c.FieldByName('RDB$DEFAULT_SOURCE').AsString);
          ffields[n].Column := ffields[n].Column + ' ' + ffields[n].DefaultValue;
          if pos('default',ffields[n].DefaultValue)=1 then
            delete(ffields[n].DefaultValue,1,Length('default'));
          ffields[n].DefaultValue := Trim(ffields[n].DefaultValue);
          if ffields[n].DefaultValue<>'' then
          if (ffields[n].DefaultValue[1]= '''') and  (ffields[n].DefaultValue[Length(ffields[n].DefaultValue)] =  '''') then
          begin
            delete(ffields[n].DefaultValue,1,1);
            delete(ffields[n].DefaultValue,Length(ffields[n].DefaultValue),1);
          end;
        end;


        { The null flag is either 1 or null (for nullable) .  if there is
          a constraint name, print that too.  Domains cannot have named
          constraints.  The column name is in rdb$trigger_name in
          rdb$check_constraints.  We hope we get at most one row back. }

        if c.FieldByName('RDB$NULL_FLAG').AsInteger = 1 then
        begin
          qryConstraints.Params.ByName('FIELDNAME').AsString := c.FieldByName('RDB$FIELD_NAME').AsString;
          qryConstraints.Params.ByName('RELATIONNAME').AsString := c.FieldByName('RDB$RELATION_NAME').AsString;
          qryConstraints.Open;

          while not qryConstraints.Eof do
          begin
            if Pos('INTEG', qryConstraints.FieldByName('RDB$CONSTRAINT_NAME').AsString) <> 1 then
              ffields[n].Column := ffields[n].Column + Format(' CONSTRAINT %s',
                [ QuoteIdentifier( FIB_Connection.SQLDialect,
                      Trim(qryConstraints.FieldByName('RDB$CONSTRAINT_NAME').AsString))]);
            qryConstraints.Next;
          end;
          qryConstraints.Close;
          ffields[n].NotNull := true;
          ffields[n].Column := ffields[n].Column + ' NOT NULL';
        end;

        if ((ffields[n].FieldType in [blr_text, blr_varying]) or
            (ffields[n].FieldType = blr_blob)) and
           (not c.FieldByName('RDB$CHARACTER_SET_ID').IsNull) and
           (c.FieldByName('RDB$CHARACTER_SET_ID').AsInteger <> 0) and
           (intchar <> 0) then
        begin
          ffields[n].CollationId := 0;
          if not c.FieldByName('RDB$FIELDS.RDB$COLLATION_ID').IsNull then
            ffields[n].CollationId := c.FieldByName('RDB$FIELDS.RDB$COLLATION_ID').AsInteger
          else
            if not c.FieldByName('RDB$RELATION_FIELDS.RDB$COLLATION_ID').IsNull then
              ffields[n].CollationId := c.FieldByName('RDB$RELATION_FIELDS.RDB$COLLATION_ID').AsInteger;

          ffields[n].CharSetId := 0;
          if not c.FieldByName('RDB$FIELDS.RDB$CHARACTER_SET_ID').IsNull then
            ffields[n].CharSetId := c.FieldByName('RDB$CHARACTER_SET_ID').AsInteger;

          if ffields[n].CollationId <> 0 then
          begin
            ffields[n].Collation := GetCollation(ffields[n].CollationId);
            ffields[n].CharacterSet := GetCharacterSets(ffields[n].CharSetId, ffields[n].CollationId, true);
            ffields[n].Column := ffields[n].Column + ffields[n].CharacterSet;
          end;
        end;
      end;

    end;

    procedure SearchField;
    var n:integer;
    begin
      for n:= low(ffields) to high(ffields) do
      begin
        if ffields[n].Name = c.Fields[0].AsString then
        begin
          AssignField(n);
          break;
        end;
      end;
    end;

var
    n:integer;
    nUIdx:string;
begin
  if Assigned(ib_connection) and (TableName<>'') then
  begin
    c:=TIB_Cursor.Create(nil);
    t:=TIB_Transaction.Create(nil);
    qryPrecision := TIB_Cursor.Create(nil);
    qryConstraints := TIB_Cursor.Create(nil);
    try
      t.IB_Connection := FIB_Connection;
      SetLength(fJoinFields,0);
      FreeConstraints;
      GetReferentialConstList(fConstraints,TableName);
      qryPrecision.IB_Connection := FIB_Connection;
      qryPrecision.IB_Transaction := t;
      qryConstraints.IB_Connection := FIB_Connection;
      qryConstraints.IB_Transaction := t;
      qryPrecision.SQL.Text :=(PrecisionSQL);
      qryPrecision.Prepare;
      qryConstraints.SQL.Text :=(ConstraintSQL);
      qryConstraints.Prepare;
      t.IB_Connection := IB_Connection;
      c.IB_Connection := IB_Connection;
      c.IB_Transaction := t;
      if IB_Connection.SQLDialect<3 then
        c.SQL.Text := TableListSQL1
      else
        c.SQL.Text := TableListSQL3;
      c.Prepare;
      c.Params[0].AsString := TableName;
      c.Open;
      if fAssignedFields and (Length(ffields)>0) then
      begin
        while not c.eof do
        begin
          SearchField;
          c.Next;
        end;
      end
      else
      begin
        n:=0;
        while not c.eof do
        begin
          if n>high(ffields) then setlength(ffields,n+1);
          AssignField(n);
          inc(n);
          c.Next;
        end;
        setlength(ffields,n);
      end;
      c.Close;

      c.SQL.Text := SqlPK;
      c.Prepare;
      c.Params[0].AsString := TableName;
      c.Open;
      fPk.Clear;
      while not c.eof do
      begin
        fPk.Add(c.Fields[0].AsString);
        c.Next;
      end;
      c.close;
      if fPk.Count = 0 then
      begin
        fSimulatePK := False;
        c.SQL.Text := SqlUIDX;
        c.Prepare;
        c.Params[0].AsString := TableName;
        c.Open;
        if not c.eof then
        begin
          nUIdx := c.Fields[0].AsString;
          c.close;
          c.SQL.Text := SqlFldIDX;
          c.Prepare;
          c.Params[0].AsString := nUIdx;
          c.Open;
          fPk.Clear;
          while not c.eof do
          begin
            fSimulatePK := True;
            fPk.Add(c.Fields[0].AsString);
            c.Next;
          end;
        end;
      end;
      t.Commit;
      for n := Low(fields) to  high(Fields) do
      begin
        fFields[n].pk := fPk.IndexOf(fFields[n].Name) >= 0;
      end;
    finally
      c.free;
      t.Free;
      qryPrecision.Free;
      qryConstraints.Free;
    end;
  end;
end;

procedure TSqlStatements.GetIndexData(var Data: TIndexDef;  const IndexName: string);
const
  SIndexList =
    ' SELECT RDB$INDEX_NAME '+
    ' ,RDB$UNIQUE_FLAG '+
    ' ,RDB$INDEX_TYPE '+
    ' ,RDB$INDEX_INACTIVE ' +
    ' ,RDB$RELATION_NAME '+
    ' ,RDB$DESCRIPTION '+
    ' FROM RDB$INDICES '+
    ' WHERE RDB$INDEX_NAME = %s ';

//Lorenzo
  SIndexFieldList3 =
    ' SELECT "RDB$FIELD_NAME" FROM "RDB$INDEX_SEGMENTS" '+
    ' WHERE "RDB$INDEX_NAME" = %s '+
    ' ORDER BY "RDB$FIELD_POSITION";';

  SIndexFieldList1 =
    ' SELECT RDB$FIELD_NAME FROM RDB$INDEX_SEGMENTS '+
    ' WHERE RDB$INDEX_NAME = %s '+
    ' ORDER BY RDB$FIELD_POSITION ';

var
  n:integer;
  SqlIndexFieldList:String;
  lQry:TIB_Cursor;
  lQry2:TIB_Cursor;
begin
  lQry := TIB_Cursor.Create(nil);
  lQry2 := TIB_Cursor.Create(nil);
  try
    lQry.IB_Connection := FIB_Connection;
    lQry.IB_Transaction := FIB_Connection.DefaultTransaction;
    lQry.SQL.Text := Format(SIndexList, [AnsiQuotedStr(IndexName,'''')]);
    lQry.Open;
    while not lQry.EOF do
    begin
      Data.Name := lQry.FieldbyName('RDB$INDEX_NAME').asString;
      Data.Table := lQry.FieldbyName('RDB$RELATION_NAME').asString;
      Data.Unique := lQry.FieldbyName('RDB$UNIQUE_FLAG').AsInteger;
      Data.IndexType :=  lQry.FieldbyName('RDB$INDEX_TYPE').AsInteger;
      Data.Inactive := lQry.FieldByName('RDB$INDEX_INACTIVE').AsInteger;
      Data.Description := lQry.FieldByName('RDB$DESCRIPTION').AsString;

      n:=0;
      if FIB_Connection.SQLDialect = 1 then
        SqlIndexFieldList := Format(SIndexFieldList1, [AnsiQuotedStr(IndexName,'''')])
      else
        SqlIndexFieldList := Format(SIndexFieldList3, [AnsiQuotedStr(IndexName,'''')]);
        
      lQry2.IB_Connection := FIB_Connection;
      lQry2.IB_Transaction := FIB_Connection.DefaultTransaction;
      lQry2.SQL.Text := SqlIndexFieldList;
      lQry2.Open;
      while not lQry2.EOF do
      begin
        if n>high(Data.fields) then
          SetLength(Data.fields,n+1);
        Data.fields[n] := lQry2.Fields[0].asString;
        inc(n);
        lQry2.Next;
      end;
      SetLength(Data.fields,n);
      lQry.Next;
    end;
  finally
    lQry.Free;
    lQry2.Free;
  end;
end;


procedure TSqlStatements.GetPKList(const RelationName:String; var PK:TStringDynArray);
var c:TIB_Cursor;
    n:integer;
    uiname:string;
begin
  if (RelationName<>'') then
  begin
    c:=TIB_Cursor.Create(nil);
    try
      c.IB_Connection := FIB_Connection;
      c.IB_Transaction := FIB_Connection.DefaultTransaction;
      c.SQL.Text := SqlPK;
      c.Prepare;
      c.Params[0].AsString := RelationName;
      c.Open;
      n:=0;
      while not c.eof do
      begin
        if n>high(PK) then setlength(PK,n+1);
        PK[n]:=c.Fields[0].AsString;
        inc(n);
        c.Next;
      end;
      setlength(PK,n);
      c.close;
      if n = 0 then
      begin
        c.SQL.Text := SqlUIDX;
        c.Prepare;
        c.Params[0].AsString := RelationName;
        c.Open;
        uiname := c.Fields[0].AsString;
        c.Close;
        if uiname<>'' then
        begin
          fSimulatePK := true;
          c.SQL.Text := SqlFldIDX;
          c.Prepare;
          c.Params[0].AsString := uiname;
          c.Open;
          n:=0;
          while not c.eof do
          begin
            if n>high(PK) then setlength(PK,n+1);
            PK[n]:=c.Fields[0].AsString;
            inc(n);
            c.Next;
          end;
          setlength(PK,n);
        end;
      end;
    finally
      c.free;
    end;
  end;
end;

procedure TSqlStatements.GetReferentialConstList(var ObjectList: TRefConstraintDefArray;  const TableName: string);
var lQry:TIB_Cursor;
const
  SReferentialList =
    'SELECT '+
    ' A.RDB$CONSTRAINT_NAME,'+
    ' A.RDB$DEFERRABLE,'+
    ' A.RDB$INITIALLY_DEFERRED, ' +
    ' B.RDB$MATCH_OPTION,'+
    ' B.RDB$UPDATE_RULE,'+
    ' B.RDB$DELETE_RULE,'+
    ' A.RDB$INDEX_NAME,'+
    ' C.RDB$RELATION_NAME ' +
    'FROM '+
    ' RDB$RELATION_CONSTRAINTS A, '+
    ' RDB$REF_CONSTRAINTS B, '+
    ' RDB$RELATION_CONSTRAINTS C ' +
    'WHERE '+
    ' A.RDB$CONSTRAINT_NAME = B.RDB$CONSTRAINT_NAME ' +
    ' AND C.RDB$CONSTRAINT_NAME = B.RDB$CONST_NAME_UQ ' +
    ' AND A.RDB$RELATION_NAME = ''%s'' ' +
    ' AND '+
//    '('+
//    ' A.RDB$CONSTRAINT_TYPE = ''PRIMARY KEY'' '+
//    ' OR '+
    ' A.RDB$CONSTRAINT_TYPE = ''FOREIGN KEY'''+
//    ') ' +
    'ORDER BY A.RDB$CONSTRAINT_NAME';

var n:integer;

begin
  lQry := TIB_Cursor.Create(nil);
  try
    lQry.IB_Connection := FIB_Connection;
    lQry.IB_Transaction := FIB_Connection.DefaultTransaction;
    lQry.SQL.Text := Format(SReferentialList, [TableName]);
    lQry.Open;
    n:=0;
    while not lQry.EOF do
    begin
      SetLength(ObjectList,n+1);
      ObjectList[n].Name :=              lQry.Fields[0].AsString;
      ObjectList[n].Deferrable :=        lQry.Fields[1].AsString;
      ObjectList[n].InitiallyDeferred := lQry.Fields[2].AsString;
      ObjectList[n].Match_Option :=      lQry.Fields[3].AsString;
      ObjectList[n].Update_Rule :=       lQry.Fields[4].AsString;
      ObjectList[n].DeleteRule :=        lQry.Fields[5].AsString;
      ObjectList[n].IndexName :=         lQry.Fields[6].AsString;
      ObjectList[n].RelationName :=      lQry.Fields[7].AsString;
      GetIndexData(ObjectList[n].IndexDef,ObjectList[n].IndexName);
      GetPKList(ObjectList[n].RelationName,ObjectList[n].rPK);
      lQry.Next;
      inc(n);
    end;
  finally
    lQry.Free;
  end;
end;

function TSqlStatements.GetInsertSQL: TStrings;
begin
  Result := fInsertSQL;
end;

function TSqlStatements.GetPK: TStrings;
begin
  result := fPk;
end;

function TSqlStatements.GetSelectSQL: TStrings;
begin
  Result := fSelectSQL;
end;

function TSqlStatements.GetUpdateSQL: TStrings;
begin
  Result := fUpdateSQL;
end;

procedure TSqlStatements.PrepareIBOQuery(q: TIBOQuery;
  const aTableName: String);
var n:integer;
begin
  q.Close;
  fAssignedFields := False;
  IB_Connection := q.IB_Connection;
  TableName := aTableName;
  GenerateSQL;
  q.SQL.Text := SelectSQL.Text;
  if fSimulatePK then
  begin
    q.KeyLinksAutoDefine := False;
    q.KeyLinks.Clear;
    for n:=0 to Pred(fpk.Count) do
    begin
      q.KeyLinks.Add(IB_Connection.mkIdent(fpk[n]));
      q.ColumnAttributes.LinkParamIsSet[ fpk[n], IB_BLANKISNULL ] := true;
    end;
  end;
  q.EditSQL.Text := UpdateSql.Text;
  q.DeleteSQL.Text := DeleteSQL.Text;
  q.InsertSQL.Text := InsertSQL.Text;
end;

procedure TSqlStatements.PrepareIBOQuery(q:TIBOQuery;const aTableName: String; aFields:array of string);
var N:integer;
begin
  q.Close;
  fAssignedFields := True;
  SetLength(ffields,Length(aFields));
  for n:= Low(aFields) to High(aFields) do
  begin
    ffields[n].Name := afields[n];
    ffields[n].Computed := False;
    ffields[n].Select := true;
  end;
  IB_Connection := q.IB_Connection;
  TableName := aTableName;
  GenerateSQL;
  q.SQL.Text := SelectSQL.Text;
  if fSimulatePK then
  begin
    q.KeyLinksAutoDefine := False;
    q.KeyLinks.Clear;
    for n:=0 to Pred(fpk.Count) do
    begin
      q.KeyLinks.Add(IB_Connection.mkIdent(fpk[n]));
      q.ColumnAttributes.LinkParamIsSet[ fpk[n], IB_BLANKISNULL ] := true;
    end;
  end;
  q.EditSQL.Text := UpdateSql.Text;
  q.DeleteSQL.Text := DeleteSQL.Text;
  q.InsertSQL.Text := InsertSQL.Text;
end;

procedure TSqlStatements.SelectField(aField: String; State: boolean);
var j:integer;
begin
  for j:=Low(fFields) to High(fFields) do
    if fFields[j].Name = aField then
    begin
      fFields[j].Select := State;
      exit;
    end;
end;

procedure TSqlStatements.SetDeleteSQL(const Value: TStrings);
begin
  FDeleteSQL.Assign(Value);
end;

procedure TSqlStatements.SetIB_Connection(const Value: TIB_Connection);
begin
  FIB_Connection := Value;
  GetFieldLists;
end;

procedure TSqlStatements.SetInsertSQL(const Value: TStrings);
begin
  FInsertSQL.Assign(Value);
end;

procedure TSqlStatements.SetSelectSQL(const Value: TStrings);
begin
  FSelectSQL.Assign(Value);
end;

procedure TSqlStatements.SetSelect_In_Pk_Order(const Value: boolean);
begin
  FSelect_In_Pk_Order := Value;
end;

procedure TSqlStatements.SetTableName(const Value: String);
begin
  FTableName := Value;
  GetFieldLists;
end;

procedure TSqlStatements.SetUpdateSQL(const Value: TStrings);
begin
  FUpdateSQL.Assign(Value);
end;


procedure TSqlStatements.SetUse_OLD_In_Pk(const Value: boolean);
begin
  FUse_OLD_In_Pk := Value;
end;

procedure TSqlStatements.SetView(const Value: boolean);
begin
  FView := Value;
end;

function TSqlStatements.GetFieldData(const aName: String): TFieldRec;
var n:Integer;
begin
  FillChar(Result,sizeof(result),#0);
  for n:= Low(fFields) to High(fFields) do
    if ffields[n].Name = aName then
    begin
      result := ffields[n];
      break;
    end;
end;

function TSqlStatements.GetFieldDataPK(const index: integer): TFieldRec;
var n:Integer;
begin
  FillChar(Result,sizeof(result),#0);
  if index in [0..fPk.count-1] then
  for n:= Low(fFields) to High(fFields) do
    if ffields[n].Name = fPk[index] then
    begin
      result := ffields[n];
      break;
    end;
end;


function TSqlStatements.GetDefaultValue(const index: integer): string;
begin
  if index in [Low(fFields) .. High(fFields)] then
  case Fields[index].FieldType of
    blr_short    :{SMALLINT}  begin Result := Fields[index].DefaultValue;    end;
    blr_long     :{INTEGER}   begin Result := Fields[index].DefaultValue;    end;
    blr_quad     :{QUAD}      begin Result := Fields[index].DefaultValue;    end;
    blr_float    :{FLOAT}     begin Result := Fields[index].DefaultValue;    end;
    blr_text     :{CHAR}      begin Result := ''''+Fields[index].DefaultValue+''''; end;
    blr_double   :{DOUBLE PRECISION} begin Result := Fields[index].DefaultValue;end;
    blr_varying  :{VARCHAR}   begin Result := ''''+Fields[index].DefaultValue+''''; end;
    blr_cstring  :{CSTRING}   begin Result := ''''+Fields[index].DefaultValue+''''; end;
    blr_blob_id  :{BLOB_ID}   begin Result := Fields[index].DefaultValue;    end;
    blr_blob     :{BLOB}      begin Result := ''''+Fields[index].DefaultValue+''''; end;
    blr_sql_time :{TIME}      begin Result := Fields[index].DefaultValue;    end;
    blr_sql_date :{DATE}      begin Result := Fields[index].DefaultValue;    end;
    blr_timestamp:{TIMESTAMP} begin Result := Fields[index].DefaultValue;    end;
    blr_int64    :{INT64}     begin Result := Fields[index].DefaultValue;    end;
    else                      begin Result := ''''+Fields[index].DefaultValue+''''; end;
  end;
end;

procedure TSqlStatements.SetJoin_FK_Fields(const Value: boolean);
begin
  FJoin_FK_Fields := Value;
end;

function TSqlStatements.QuoteId(const Value:String): String;
begin
  Result := QuoteIdentifier(IB_Connection.SQLDialect,Value);
end;

function TSqlStatements.QuoteIdTF(const Table,Field:String): String;
begin
  Result := QuoteIdentifier(IB_Connection.SQLDialect,Table) +'.'+
            QuoteIdentifier(IB_Connection.SQLDialect,Field);
end;

function TSqlStatements.HasJoin: boolean;
begin
  result := Length(fJoinFields)>0;
end;

function TSqlStatements.FieldHasJoin(const FieldName: string): boolean;
var
  j:integer;
begin
  Result:=False;
  for j:=Low(fJoinFields) to High(fJoinFields) do
    if fJoinFields[j].OnLocField = FieldName then
    begin
      Result:=true;
      Break;
    end;
end;

end.
